@@ -1,7 +1,9 @@
 #include "sbox.h"
 
+#include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <libgen.h>
 #include <stdarg.h>
 #include <string.h>
 #include <strings.h>
@@ -166,6 +168,8 @@ typedef struct{
 	int do_suid;
 	int do_sgid;
 
+	int do_log_envvars;
+
 	int allow_web_owned_scripts;
 
 	int uid_min;
@@ -181,7 +185,7 @@ typedef struct{
 	int use_safe_path;
 	char* log_file;
 	char* root;
-	char* cgi_bin;	
+	char* cgi_bin;
 
 	int set_limits;
 	int priority;
@@ -201,6 +205,11 @@ typedef struct{
 	int limit_nproc_soft;
 	int limit_nofile_hard;
 	int limit_nofile_soft;
+
+	char* php_path;
+	char* python_path;
+	char* perl_path;
+	char* ruby_path;
 }sbox_config_t;
 sbox_config_t sbox_config;
 
@@ -208,6 +217,7 @@ DOTCONF_CB(cb_web_user);
 DOTCONF_CB(cb_web_group);
 DOTCONF_CB(cb_do_suid);
 DOTCONF_CB(cb_do_sgid);
+DOTCONF_CB(cb_do_log_envvars);
 DOTCONF_CB(cb_allow_web_owned_scripts);
 DOTCONF_CB(cb_uid_min);
 DOTCONF_CB(cb_uid_max);
@@ -239,12 +249,17 @@ DOTCONF_CB(cb_limit_nproc_hard);
 DOTCONF_CB(cb_limit_nproc_soft);
 DOTCONF_CB(cb_limit_nofile_hard);
 DOTCONF_CB(cb_limit_nofile_soft);
+DOTCONF_CB(cb_php_path);
+DOTCONF_CB(cb_python_path);
+DOTCONF_CB(cb_perl_path);
+DOTCONF_CB(cb_ruby_path);
 
 static configoption_t options[] = {
 	{"web_user",                ARG_STR,    cb_web_user, NULL, 0},
 	{"web_group",               ARG_STR,    cb_web_group, NULL, 0},
 	{"do_suid",                 ARG_TOGGLE, cb_do_suid, 0, 0},
 	{"do_sgid",                 ARG_TOGGLE, cb_do_sgid, 0, 0},
+	{"do_log_envvars",          ARG_TOGGLE, cb_do_log_envvars, 0, 0},
 	{"allow_web_owned_scripts", ARG_TOGGLE, cb_allow_web_owned_scripts, 0, 0},
 	{"uid_min",                 ARG_INT,    cb_uid_min, 0, 0},
 	{"uid_max",                 ARG_INT,    cb_uid_max, 0, 0},
@@ -276,6 +291,10 @@ static configoption_t options[] = {
 	{"limit_nproc_soft",        ARG_INT,    cb_limit_nproc_soft, 0, 0},
 	{"limit_nofile_hard",       ARG_INT,    cb_limit_nofile_hard, 0, 0},
 	{"limit_nofile_soft",       ARG_INT,    cb_limit_nofile_soft, 0, 0},
+	{"php_path",                ARG_STR,    cb_php_path, NULL, 0},
+	{"python_path",             ARG_STR,    cb_python_path, NULL, 0},
+	{"perl_path",               ARG_STR,    cb_perl_path, NULL, 0},
+	{"ruby_path",               ARG_STR,    cb_ruby_path, NULL, 0},
 	LAST_OPTION
 };
 DOTCONF_CB(cb_web_user){
@@ -296,6 +315,10 @@ DOTCONF_CB(cb_do_sgid){
 	sbox_config.do_sgid = cmd->data.value;
 	return NULL;
 }
+DOTCONF_CB(cb_do_log_envvars){
+	sbox_config.do_log_envvars = cmd->data.value;
+	return NULL;
+}
 DOTCONF_CB(cb_allow_web_owned_scripts){
 	sbox_config.allow_web_owned_scripts = cmd->data.value;
 	return NULL;
@@ -425,10 +448,30 @@ DOTCONF_CB(cb_limit_nofile_soft){
 	sbox_config.limit_nofile_soft = cmd->data.value;
 	return NULL;
 }
+DOTCONF_CB(cb_php_path){
+	sbox_config.php_path = (char*)malloc(strlen(cmd->data.str)+1);
+	strcpy(sbox_config.php_path,cmd->data.str);
+	return NULL;
+}
+DOTCONF_CB(cb_python_path){
+	sbox_config.python_path = (char*)malloc(strlen(cmd->data.str)+1);
+	strcpy(sbox_config.python_path,cmd->data.str);
+	return NULL;
+}
+DOTCONF_CB(cb_perl_path){
+	sbox_config.perl_path = (char*)malloc(strlen(cmd->data.str)+1);
+	strcpy(sbox_config.perl_path,cmd->data.str);
+	return NULL;
+}
+DOTCONF_CB(cb_ruby_path){
+	sbox_config.ruby_path = (char*)malloc(strlen(cmd->data.str)+1);
+	strcpy(sbox_config.ruby_path,cmd->data.str);
+	return NULL;
+}
 
-void read_config_file(){
+void read_config_file(char* sboxconf_path){
 	configfile_t *configfile;
-	configfile = dotconf_create("/etc/sbox.conf", options, 0, CASE_INSENSITIVE);
+	configfile = dotconf_create(sboxconf_path, options, 0, CASE_INSENSITIVE);
 	if (dotconf_command_loop(configfile) == 0){
 		fprintf(stderr, "Error reading config file\n");
 		exit(2);
@@ -437,14 +480,200 @@ void read_config_file(){
 	return;
 }
 
+// Maxmimum string size for an environment value:
+#define MAXENV 1024
+
+// Get the HTTP_HOST environment variable value
+// Returns NULL if not found
+char* get_http_host(char** e){
+	char *name,*delimiter;
+	static char envvar_name[MAXENV+2];      // Stores the current env var name, make sure it's big enough by adding 2 safety chars
+	while ((name = *e++)) {  
+		delimiter = (char*) strchr(name,'=');
+		if (delimiter == NULL) continue;        // If we don't find the "=" delimiter, then we didn't find anything...
+		if (delimiter - name >= MAXENV) continue;       // If our envvar name is too big, then it's as if we didn't find anything
+		strncpy(envvar_name,name,delimiter-name);        // Copy the envvar name in a buffer, to check for it later
+		envvar_name[delimiter-name]='\0';                // Make sure it terminates by a delimiting zero
+		delimiter++;                            // Point just right after the = sign
+		if( strcmp(envvar_name,"HTTP_HOST") == 0){       // Have we found the HTTP_HOST?
+			// log_error("%s => %s\n",envvar_name,delimiter);   // Log the HTTP_HOST in our sbox.log
+			return delimiter;
+		}
+	}
+	return NULL;
+}
+
+// Debug function to print the environment values in the sbox.log
+// Do not activate all the time, as its printing quite some stuff,
+// and flooding a logfile is never a good idea.
+void print_the_env(char** e){
+	char *name,*delimiter;
+	static char nm[MAXENV+2];
+	while ((name = *e++)) {  
+		delimiter = (char*) strchr(name,'=');
+		if (delimiter == NULL) continue;
+		if (delimiter - name >= MAXENV) continue;
+		strncpy(nm,name,delimiter-name);
+		nm[delimiter-name]='\0';
+		log_error("%s => %s\n",nm,++delimiter);
+	}
+}
+
+int isToggleValueOn(char* t){
+	if( strcmp(t,"On") == 0 || strcmp(t,"on") == 0 || strcmp(t,"1") == 0 || strcmp(t,"ON") == 0)
+		return 1;
+	else
+		return 0;
+}
+// This function will read the environment variables of
+// Apache, so that you can eventually set the configuration
+// of SBOX in it in order to overwrite what is in /etc/sbox.conf
+//
+// This gives the additional functionality to be able to have
+// specific values depending on which vhost you run.
+//
+// For example, in your apache config file, you can do:
+// SetEnv SBOX_DTC_CONF_web_user 100
+// in order to overwrite the "web_user" value.
+// Of course, you will need the "mod_env" apache module to do that.
+void readEnvSboxConfig(char** e){
+	char *name,*delimiter;
+	static char nm[MAXENV];
+	while ((name = *e++)) {
+		delimiter = (char*) strchr(name,'=');
+		if (delimiter == 0) continue;
+		if (delimiter - name >= MAXENV) continue;
+		strncpy(nm,name,delimiter-name);
+		nm[delimiter-name]='\0';
+		delimiter++;
+		if( strcmp(nm,"SBOX_DTC_CONF_web_user") == 0)			sbox_config.web_user = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_web_group") == 0)			sbox_config.web_group = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_allow_web_owned_scripts") == 0)	sbox_config.allow_web_owned_scripts = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_uid_min") == 0)			sbox_config.uid_min = atoi(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_uid_max") == 0)			sbox_config.uid_max = atoi(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_gid_min") == 0)			sbox_config.gid_min = atoi(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_gid_max") == 0)			sbox_config.gid_max = atoi(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_sid_mode_directory") == 0)		sbox_config.sid_mode_directory = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_use_safe_path") == 0)		sbox_config.use_safe_path = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_safe_path") == 0)			sbox_config.safe_path = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_log_file") == 0)			sbox_config.log_file = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_do_log_envvars") == 0)		sbox_config.do_log_envvars = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_echo_failures") == 0)		sbox_config.echo_failures = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_do_chroot") == 0)			sbox_config.do_chroot = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_root") == 0)			sbox_config.root = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_cgi_bin") == 0)			sbox_config.cgi_bin = delimiter;
+		if( strcmp(nm,"SBOX_DTC_CONF_do_suid") == 0)			sbox_config.do_suid = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_do_sgid") == 0)			sbox_config.do_sgid = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_set_limits") == 0)			sbox_config.set_limits = isToggleValueOn(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_priority") == 0)			sbox_config.priority = atoi(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_cpu_hard") == 0)		sbox_config.limit_cpu_hard = atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_cpu_soft") == 0)		sbox_config.limit_cpu_soft = atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_fsize_hard") == 0)		sbox_config.limit_fsize_hard = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_fsize_soft") == 0)		sbox_config.limit_fsize_soft = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_data_hard") == 0)		sbox_config.limit_data_hard = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_data_soft") == 0)		sbox_config.limit_data_soft = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_stack_hard") == 0)			sbox_config.limit_stack_hard = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_stack_soft") == 0)		sbox_config.limit_stack_soft = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_core_hard") == 0)		sbox_config.limit_core_hard = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_core_soft") == 0)		sbox_config.limit_core_soft = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_rss_hard") == 0)		sbox_config.limit_rss_hard = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_rss_soft") == 0)		sbox_config.limit_rss_soft = 1024*atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_nproc_hard") == 0)		sbox_config.limit_nproc_hard = atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_nproc_soft") == 0)		sbox_config.limit_nproc_soft = atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_nofile_hard") == 0)		sbox_config.limit_nofile_hard = atol(delimiter);
+		if( strcmp(nm,"SBOX_DTC_CONF_limit_nofile_soft") == 0)		sbox_config.limit_nofile_soft = atol(delimiter);
+	}
+}
+
 /*****************  START OF CODE ****************/
 
-int main (int argc, char* argv[]) {
-	struct stat target_stat,dir_stat;
-	char *target,*newroot,*root,*additional_info,*cmd;
-	int gid,uid,i,return_int;
+char* getFileExtension(char* file){
+	int len,i;
 
-	read_config_file();
+	len = strlen(file);
+	for(i=len-1 ; i>0 ; --i){
+		if( file[i] == '.'){
+			return &file[i];
+		}
+	}
+	return 0;
+}
+
+#define HOST_CONF_FILE_FOLDER "/etc/sbox/vhosts.d/"
+int main (int argc, char* argv[], char** e) {
+	struct stat target_stat,dir_stat;
+	char *target,*newroot,*root,*additional_info,*cmd,*new_cmd_interp;
+	int gid,uid,i,return_int;
+	char* my_http_host;
+	char* ext;
+	struct stat stat_buffer;
+	char* http_host_config_file_path;
+	int conf_file_path_string_size;
+	char realpath_path[PATH_MAX+2];
+	char* dirname_path;
+	char* dirname_path2;
+	FILE* fp;
+
+	////////////////////////// GET THE PATH OF THE SBOX CONFIG DEPENDING ON THE HTTP_HOST ENV VAR //////////////////////////
+	//sbox_config.log_file = "/var/log/sbox.log";
+	//open_log();
+	//log_error("Testing log\n");
+	my_http_host = get_http_host(e);	// Get the current vhost name
+	if(my_http_host == NULL){
+		sbox_config.log_file = "/var/log/sbox.log";
+		open_log();
+		log_error("HTTP_HOST environment variable not found: exiting.\n");
+		exit(1);
+	}
+	if(strlen(my_http_host) > 1024){
+		sbox_config.log_file = "/var/log/sbox.log";
+		open_log();
+		log_error("HTTP_HOST environment variable too big: exiting.\n");
+		exit(1);
+	}
+	conf_file_path_string_size = strlen(HOST_CONF_FILE_FOLDER) + strlen(my_http_host) + 2;
+	http_host_config_file_path = malloc(conf_file_path_string_size);
+	if(http_host_config_file_path == NULL){
+		sbox_config.log_file = "/var/log/sbox.log";
+		open_log();
+		log_error("Could not alloc space for storing HTTP_HOST: exiting.\n");
+		exit(1);
+	}
+	strcpy(http_host_config_file_path,HOST_CONF_FILE_FOLDER);
+	strcat(http_host_config_file_path,my_http_host);
+	// log_error("Concat file is: \"%s\"\n",http_host_config_file_path);
+	if(!realpath(http_host_config_file_path,realpath_path)){
+		// If not resolved, open the default config file
+		//log_error("realpath returned NULL: opening default /etc/sbox/sbox.conf\n");
+		read_config_file("/etc/sbox/sbox.conf");
+	}else{
+		// If resolved, then there may be a vhost specific file that can be opened!
+		log_error("Realpath is: %s\n",realpath_path);
+		dirname_path = strdup(realpath_path);
+		dirname_path2 = dirname(dirname_path);
+		//log_error("dirname_path2 is: %s\n",dirname_path2);
+		strcat(dirname_path2,"/");
+		if( strcmp(dirname_path,HOST_CONF_FILE_FOLDER) != 0 ){
+			sbox_config.log_file = "/var/log/sbox.log";
+			open_log();
+			log_error("Directory traversal attempt: exiting.\n");
+			exit(1);
+		}
+
+		// Get to know if the file exists, if it does use it as config file, if not use /etc/sbox/sbox.conf
+		fp = fopen(realpath_path,"r");
+		if(fp == NULL){
+			//log_error("File %s doesn't exist or can't be read, using /etc/sbox/sbox.conf",realpath_path);
+			read_config_file("/etc/sbox/sbox.conf");
+		}else{
+			//log_error("File exist: using %s\n",realpath_path);
+			fclose(fp);
+			read_config_file(realpath_path);
+		}
+	}
+	// Used for debug purpose: do ont use in production, as this is very easy
+	// to hack a change in the environment of Apache
+	// readEnvSboxConfig(e);
 #if defined(DEBUG)
 	printf("Web user: %s\n",sbox_config.web_user);
 	printf("Web group: %s\n",sbox_config.web_group);
@@ -481,7 +710,16 @@ int main (int argc, char* argv[]) {
 	printf("limit_nofile_hard: %d\n",sbox_config.limit_nofile_hard);
 	printf("limit_nofile_soft: %d\n",sbox_config.limit_nofile_soft);
 #endif
+
+	if(sbox_config.do_log_envvars){
+		print_the_env(e);
+	}
 	open_log();
+
+	if(sbox_config.do_log_envvars){
+		print_the_env(e);
+	}
+
 	check_consistency();
 
 	/* figure out the base of the user's HTML document tree */
@@ -525,6 +763,13 @@ int main (int argc, char* argv[]) {
 
 	clean_env();
 
+	/* set hard and soft limits on resource usage */
+	/* important to do this after log is closed because otherwise the log file 
+	  size will be restricted by the LIMIT_FSIZE_SOFT value */
+	if(sbox_config.set_limits)
+		set_limits();
+
+
 	/* set to the owner's uid and gid */
 	if(sbox_config.do_suid || sbox_config.do_sgid){
 		gid = sbox_config.sid_mode_directory == 1 ? dir_stat.st_gid : target_stat.st_gid;
@@ -563,26 +808,44 @@ int main (int argc, char* argv[]) {
 	if(strlen(sbox_config.log_file) > 0){
 		log_error("executing %s for uid %d, gid %d\n",target,getuid(),getgid());
 	}
-	/* set hard and soft limits on resource usage */
-	/* important to do this after log is closed because otherwise the log file 
-	  size will be restricted by the LIMIT_FSIZE_SOFT value */
-	if(sbox_config.set_limits)
-		set_limits();
 
-	/* fix up argv[] array */
-	argv[0] = cmd;
+
+	/* fix up argv[] array and command depending on the extension */
+	ext = getFileExtension(cmd);
+	if(ext == 0){
+		new_cmd_interp = cmd;
+		argv[0] = cmd;
+		log_error("No extension found: starting as CGI\n");
+	}else{
+		if( strcmp(ext,".php") == 0){
+			new_cmd_interp = sbox_config.php_path;
+		}else if(strcmp(ext,".py") == 0){
+			new_cmd_interp = sbox_config.python_path;
+		}else if(strcmp(ext,".pl") == 0){
+			new_cmd_interp = sbox_config.perl_path;
+		}else if(strcmp(ext,".rb") == 0){
+			new_cmd_interp = sbox_config.ruby_path;
+		}else{
+			log_error("Extension not recognised: will exit here!\n");
+			return -1;
+		}
+		argv[0] = new_cmd_interp;
+		argv[1] = cmd;
+		log_error("Extension is: %s, will use %s as interpreter\n",ext,new_cmd_interp);
+	}
 
 	/* chdir() to new directory */
 	chdir_fix(newroot,target);
 	if(strlen(sbox_config.log_file) > 0){
-		log_error("Going to try this: %s , %s, %s\n", cmd, argv[1], cleanenv[0]);
+		log_error("Going to try this: %s , %s, %s\n", new_cmd_interp, argv[1], cleanenv[0]);
 	}
 
+
 //if (log && log != stderr)
 //    fclose(log);
 
 	/* execute the target */
-	return_int = execve(cmd,argv,cleanenv);
+	return_int = execve(new_cmd_interp,argv,cleanenv);
 
 	/* if we get here, then something went wrong -- we won't be able to
 	   reopen the log file because we've given up privileges, logging
@@ -590,7 +853,7 @@ int main (int argc, char* argv[]) {
 	   it's expecting a Content-Type header, so disable logging via
 	   write_log entirely by setting log to NULL. */
 	log = NULL;
-	fatal_error("exec of %s failed (%s)\n",escape(cmd),strerror(errno));
+	fatal_error("exec of %s failed (%s)\n",escape(new_cmd_interp),strerror(errno));
 
 	return -1;
 }
@@ -776,6 +1039,7 @@ char* fetch_target(char* root,char** additional,struct stat* starg,struct stat*
 	   or zero length */
 	path_translated = (char*) getenv("PATH_TRANSLATED");
 	path_info = (char*) getenv("PATH_INFO");
+	// log_error("path_info %s , path_translated %s\n", path_info, path_translated );
 
 	if (path_info[0] != '/')
 		fatal_error("PATH_INFO does not begin with a '/'\n");
@@ -786,6 +1050,10 @@ char* fetch_target(char* root,char** additional,struct stat* starg,struct stat*
 		shaved = shave(root,path_translated);
 	else
 		shaved = path_info;
+	
+	//log_error("shaved %s\n", shaved);
+	//log_error("root %s\n", root);
+	//log_error("sbox_config.cgi_bin %s\n", sbox_config.cgi_bin);
 
 	if (strlen(root)+1+strlen(sbox_config.cgi_bin) > MAXPATHLEN)
 		fatal_error("Script directory path larger than MAXPATHLEN\n");
